import numpy as np
import uncertainties.unumpy as unp
from uncertainties import ufloat, UFloat
import numbers  # to test if something is a number


def __parseWord__(word):
    """Parse a word from a tablefile"""

    word = word.strip()
    if len(word) == 0:
        return None
    if word[-1] == '"' and word[0] == '"':
        return word[1:-1]

    try:
        ret = float(word)
        return float(ret)
    except:
        return word


def genfromtxt(filename):
    """Generates a Tablehandler out of the table in a given file"""

    # Open file and get lines
    with open(filename) as f:
        lines = f.readlines()
        f.close()

    # Parse table
    table = []
    columns = []
    for line in lines:
        line = line.rstrip()

        indexH = line.find("#")
        if indexH != -1:
            line = line[:indexH]

        # Determine position of columns
        if len(columns) == 0:
            if line.strip() == "":
                continue

            columns = [0]
            foundFirstArg = False
            inArg = False
            inQuotes = False

            for i in range(len(line)):
                if line[i] == '#' or line[i] == "\n" or line[i] == "\f":
                    break

                if foundFirstArg == False:
                    if line[i] != " ":
                        foundFirstArg = True
                        inArg = True
                        if line[i] == '"':
                            inQuotes = True
                elif inQuotes:
                    if line[i] == '"':
                        inQuotes = False
                        inArg = False
                elif inArg == False and line[i] != " ":
                    columns.append(i)
                    inArg = True
                    if line[i] == '"':
                        inQuotes = True
                elif inArg and line[i] == " ":
                    inArg = False

            for i in columns:
                table.append(list())

        found_last = False

        for i in range(len(columns)-1):

            if(len(line) <= columns[i+1]):
                newWord = __parseWord__(line[columns[i]:])
                table[i].append(newWord)
                found_last = True

                # Fill table with nones so we get a rectangular shape
                for j in range(i+1, len(columns)):
                    table[j].append(None)
                break

            newWord = __parseWord__(line[columns[i]:columns[i+1]])
            table[i].append(newWord)

        if not found_last:
            newWord = __parseWord__(line[columns[-1]:])
            table[-1].append(newWord)

    return TableHandler(table)


class TableHandler:
    """ Class to handle tables in the internship easily """

    def __init__(self, table=None):
        """Initalize table, eather from given table or from Filename"""
        self.table = table

    def getDict(self):
        """Retuns a dictionary mapping the first line to its columns"""
        dicti = {}
        for c in self.table:
            dicti[c[0]] = c[1:]
        return dicti

    def getTransposedTable(self, table=None):
        """Returns the transposition of the (underlaying, if table = None) table."""
        if table == None:
            table = self.table

        tTable = []
        for i in table:
            while len(i) > len(tTable):
                tTable.append([])

        for c in table:
            for i in range(len(c)):
                tTable[i].append(c[i])

        return tTable

    def transpose(self):
        """Transposes underlaying table"""
        self.table = self.getTransposedTable()

    def insertRow(self, row, index=None):
        """Add a segment to the table
        row: 1d Table to  add
        index: index to add the table to. Set to None to append"""
        if index == None:
            self.table.append(row)
        else:
            self.table.insert(index, row)

    def popRow(self, index=-1):
        """Deletes row with index index"""
        self.table.pop(index)

    def getRawDict(self):
        """Returns a dictionary mapping the first line to its non-string collumn values"""
        dicti = {}
        for c in self.table:
            dicti[c[0]] = []
            for j in c[1:]:
                if isinstance(j, numbers.Number):
                    dicti[c[0]].append(j)

        return dicti

    def getRawData(self):
        """Returns only floats and ints"""

        data = []
        for i in self.table:
            data.append([])

        for i in range(len(self.table)):
            for w in self.table[i]:
                if isinstance(w, numbers.Number):
                    data[i].append(w)
        return data

    def makeTexTable(self, filename, useSIUnitX=True, precision=None, makeHeader=True, standardRules=True):
        """Generates a .tex table into file"""

        # Generate transposed table
        tTable = self.getTransposedTable()
        # Search length and replace Nones
        maxLen = 0

        maxRowLen = 0

        # find length of row
        for i in tTable:
            if maxRowLen < len(i):
                maxRowLen = len(i)

        for i in range(len(tTable)):
            if len(tTable[i]) < maxRowLen:
                tTable[i] += (maxRowLen - len(tTable[i])) * [""]

            for j in range(len(tTable[i])):
                if(tTable[i][j] == None):
                    tTable[i][j] = ""

                if maxLen < len(str(tTable[i][j])):
                    maxLen = len(str(tTable[i][j]))

        # account for SIUnitX stuff
        maxLen += 10

        # Write TeX
        file = open(filename, "w+")

        if makeHeader:
            file.write(r"\begin{table}"+"\n")

            file.write("\t"+r"\caption{TABLE}"+"\n")
            file.write("\t"+r"\label{tab:NAME}"+"\n")
            if(useSIUnitX):
                file.write("\t"+r"\sisetup{table-format=X.")
                if(precision != None):
                    file.write(str(int(precision)))
                else:
                    file.write("X")
                file.write("}\n")

            file.write("\t"+r"\begin{tabular}{")
            if useSIUnitX:
                for i in self.table[:-1]:
                    file.write("S ")
                file.write("S")
            else:
                for i in self.table[:-1]:
                    file.write("c ")
                file.write("c")
            file.write("}\n")

            if standardRules:
                if makeHeader:
                    file.write("\t\t")
                file.write(r"\toprule"+"\n")

        def parseTexWord(word, end=False):
            newWord = ""
            if isinstance(word, numbers.Number) or isinstance(word, UFloat):
                if precision != None:
                    newWord = ("{:."+str(int(precision))+"f}").format(word)
                else:
                    newWord = str(word)
            else:
                newWord = str(word)
                if(useSIUnitX):
                    newWord = "{"+newWord+"}"

            if makeHeader:
                file.write("\t\t")

            file.write(newWord+(maxLen - len(newWord))*" ")

            if not end:
                file.write(" & ")

        for l in range(len(tTable)):

            if standardRules and l == 1:
                if makeHeader:
                    file.write("\t\t")
                file.write(r"\midrule"+"\n")

            for w in tTable[l][:-1]:
                parseTexWord(w)

            parseTexWord(tTable[l][-1], end=True)
            file.write(r"  \\"+"\n")

        if standardRules:
            if makeHeader:
                file.write("\t\t")
            file.write(r"\bottomrule"+"\n")

        if makeHeader:
            file.write("\t"+r"\end{tabular}"+"\n")
            file.write(r"\end{table}")

        file.close()

    def getMeanValues(self):
        """Returns the mean values and errors of them as ufloats"""

        rawDict = self.getRawDict()
        meanValues = {}
        for v in rawDict.items():
            meanValues[v[0]] = np.mean(ufloat(np.mean(v[1]), np.std(v[1])))

        return meanValues



# Setting matplotlib layout to match tex settings

import matplotlib
import matplotlib.pyplot as plt
from os.path import dirname, abspath
import locale

def cla():
    plt.cla()
    plt.ticklabel_format(useLocale=True)

matplotlib.use('pgf')
matplotlib.rcParams.update({
    'font.family': 'serif',
    'text.usetex': True,
    'pgf.rcfonts': False,
    'pgf.texsystem': 'lualatex',
    'pgf.preamble': r'\input{'+dirname(abspath(__file__)).replace(" ", r"\ ")+r'//matplotlib_header.tex'+r'}',
})

# use german locale settings for printing 3.4 as 3,4
try:
    locale.setlocale(locale.LC_ALL, 'de_DE.UTF8')
except locale.Error:
    print("Could not set the language settings! 3.5 will not be written as 3,5! SO SAD!")

plt.ticklabel_format(useLocale=True)



#!/usr/bin/env python3
# Parse args
from praktipy import TableHandler
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("TableFile", help='Path to the table file. E.g. "Table.txt"')
parser.add_argument("-m", "--meanValues", dest="mean", help="Print mean values", action="store_true")
parser.add_argument("-t", "--tex", dest="tex", help="Write a Tex Table to file", metavar="TEXFILE")
args = parser.parse_args()

th = TableHandler(args.TableFile)

print("Found following table:")
for i in th.table:
    print(i)
print("==============")

if args.mean:
    print("Mean values:")
    for i in th.getMeanValues().items():
        print(str(i[0])+":", i[1])
    print("==============")

if args.tex:
    th.makeTexTable(args.tex)
    print("Generated tex table as", args.tex)



# Setting matplotlib layout to match tex settings

import locale
import matplotlib
import platform
import numpy as np
import matplotlib.pyplot as plt
from os.path import dirname, abspath

# use german locale settings for printing 3.5 as 3,5
locale_strings = {
    'Darwin': 'de_DE',
    'Windows': 'de_de',
    'Linux': 'de_DE.UTF8',
}
matplotlib.use('pgf')


def auto_set_backend():
    import sys
    if "release" in sys.argv:
        print("Setting up Matplotlib to the slow, but pretty backend.")
        pretty()
    else:
        print("Setting up Matplotlib to the fast backend.")
        fast()

def fast():
    matplotlib.rcParams.update({
        'font.family': 'serif',
        'text.usetex': True,
        'pgf.rcfonts': False,
        'pgf.texsystem': 'pdflatex',
        'pgf.preamble': r'\input{'+dirname(abspath(__file__)).replace(" ", r"\ ")+r'//matplotlib_header_minimal.tex'+r'}'
    })


def pretty():  
    try:
        locale.setlocale(locale.LC_ALL, locale_strings[platform.system()])
    except locale.Error:
        print("Could not set the language settings! 3.5 will not be written as 3,5! SO SAD!")

    matplotlib.rcParams.update({
        'font.family': 'serif',
        'text.usetex': True,
        'pgf.rcfonts': False,
        'pgf.texsystem': 'lualatex',
        'pgf.preamble': r'\input{'+dirname(abspath(__file__)).replace(" ", r"\ ")+r'//matplotlib_header.tex'+r'}',
        'axes.formatter.use_locale' : True,
    })

# Currently not working
# if platform.system() == "Windows":
#     plt.style.use('file:\\' + os.path.dirname(os.path.abspath(__file__)) + os.sep + 'praktipy.mplstyle')
# else:
#     plt.style.use('file://' + os.path.dirname(os.path.abspath(__file__)) + os.sep + 'praktipy.mplstyle')

# print('file://' + os.path.dirname(os.path.abspath(__file__)) + os.sep + 'praktipy.mplstyle')
# convenience functions

def polyplotfit(x, params, N=1000, border=0.05):
    """Plots a polynome, which was fitted.
    x: the original x value which was fitted
    params: the parameters in the polynome
    N: Number of x_values to calculate
    border: percentage of x_range to make a border"""

    dx = x[-1] - x[0]
    x_fit = np.linspace(x[0] - dx*border, x[-1] + dx*border, N)
    y_fit = np.zeros(len(x_fit))
    deg = len(params)
    for i in range(deg):
        y_fit += params[deg-1-i] * x_fit**i

    return (x_fit, y_fit)


def curveplotfit(f, x, params, N=1000, border=0.05, logscale=False):
    """Plots a general function, which was fitted.
    x: the original x value which was fitted
    params: the parameters in the polynome
    N: Number of x_values to calculate
    border: percentage of x_range to make a border"""
    dx = x[-1] - x[0]
    if logscale:
        x_fit = np.logspace(np.log10(
            x[0]) - np.log10(x[0])*border*0.6, np.log10(x[-1])+np.log10(x[-1])*border*0.2, N)
    else:
        x_fit = np.linspace(x[0] - dx*border, x[-1] + dx*border, N)
    y_fit = f(x_fit, *params)

    return (x_fit, y_fit)

def cla():
    """
    Calls the matplotlib plt.cla, but turns the useLocale option back on.
    """
    plt.cla()
    plt.ticklabel_format(useLocale=True)



# to test if something is a number
import numbers
import warnings
import math
import numpy as np
from uncertainties import ufloat, UFloat
from math import log10


def gen_from_txt(filename, explicit_none=False):
    """Generates a multi-datatype table (A python list of python lists,
    which contains strings, floats and None).
    You can generate uncertaintie ufloats by writing something like
    42.000+-3.1415.

    WARNING: Not suitable for very large tables (everything under a few thousand line should be fine though).
        Use numpy.genfromtxt instead!

    Parameters
    ----------
    filename : str
        The path to the textfile to parse. Look into the examples, for further
        information on how to format the file.
    explicitNone : bool, optional
        Whether the table structure is determend visually by looking
        at the first line, or with explicit "Nones" where no value is present.
        (the default is False, which looks at the file visually)

    Returns
    -------
    list of lists (python lists) [][]
        The parsed table
    """

    if explicit_none:
        return _gen_from_txt_explicit(filename)
    else:
        return _gen_from_txt_visual(filename)


def _gen_from_txt_visual(filename):
    table = []
    column_lengths = []

    # Parse table
    with open(filename) as file:
        for line in file:
            line = line.rstrip()

            # check for comments beginning with '#'
            index_end = line.find("#")
            if index_end != -1:
                line = line[:index_end]

            # Determine position of columns
            # len(columns) == 0
            if not column_lengths:
                if line.strip() == "":
                    continue

                column_lengths = [0]
                found_first_arg = False
                in_arg = False
                in_quotes = False

                for i in range(len(line)):
                    if line[i] == '#' or line[i] == "\n" or line[i] == "\f":
                        break

                    if not found_first_arg:
                        if line[i] != " ":
                            found_first_arg = True
                            in_arg = True
                            if line[i] == '"':
                                in_quotes = True
                    elif in_quotes:
                        if line[i] == '"':
                            in_quotes = False
                            in_arg = False
                    elif not in_arg and line[i] != " ":
                        column_lengths.append(i)
                        in_arg = True
                        if line[i] == '"':
                            in_quotes = True
                    elif in_arg and line[i] == " ":
                        in_arg = False

                for i in column_lengths:
                    table.append(list())

            found_last = False
            for i in range(len(column_lengths)-1):

                if len(line) <= column_lengths[i+1]:
                    new_word = __parse_word__(line[column_lengths[i]:])
                    table[i].append(new_word)
                    found_last = True
                    # Fill table with nones so we get a rectangular shape
                    for j in range(i+1, len(column_lengths)):
                        table[j].append(None)
                    break

                new_word = __parse_word__(
                    line[column_lengths[i]:column_lengths[i+1]])
                table[i].append(new_word)

            if not found_last:
                new_word = __parse_word__(line[column_lengths[-1]:])
                table[-1].append(new_word)

        file.close()

    return table


def _gen_from_txt_explicit(filename):
    # Can you see my c++-heritage?
    class StateEnum:
        IN_WORD = 0
        IN_QUOTES = 1
        IN_BETWEEN = 2

    # Table to return
    table = []

    # Parse file
    with open(filename) as file:
        # Parse every line
        for line in file:
            # to remember column
            column = 0
            # Ignore whitespace and comments (with #)
            line = line.strip()
            index_end = line.find("#")
            if index_end != -1:
                line = line[:index_end]

            new_word = ""
            state = StateEnum.IN_BETWEEN

            for char in line:
                if state == StateEnum.IN_BETWEEN:
                    if char == " ":
                        pass
                    elif char == "\n" or char == "\f" or char == "\0":
                        break
                    elif char == '"':
                        state = StateEnum.IN_QUOTES
                        new_word += char
                    else:
                        new_word += char
                        state = StateEnum.IN_WORD
                elif state == StateEnum.IN_QUOTES:
                    if char == " ":
                        new_word += char
                    elif char == "\n" or char == "\f" or char == "\0":
                        __add_word__(__parse_word__(new_word), table, column)
                        column += 1
                        new_word = ""
                        break
                    elif char == '"':
                        new_word += char
                        __add_word__(__parse_word__(new_word), table, column)
                        column += 1
                        new_word = ""
                        state = StateEnum.IN_BETWEEN
                    else:
                        new_word += char
                elif state == StateEnum.IN_WORD:
                    if char == " ":
                        __add_word__(__parse_word__(new_word), table, column)
                        column += 1
                        new_word = ""
                        state = StateEnum.IN_BETWEEN
                    elif char == "\n" or char == "\f" or char == "\0":
                        __add_word__(__parse_word__(new_word), table, column)
                        column += 1
                        new_word = ""
                        break
                    elif char == '"':
                        new_word += char
                    else:
                        new_word += char

            if new_word != "":
                __add_word__(__parse_word__(new_word), table, column)

        file.close()
    return table


def __add_word__(word, table, column):
    if len(table) < column + 1:
        table.append([word])
        return
    table[column].append(word)


def __parse_word__(word):
    """Parse a word from a tablefile"""

    word = word.strip()
    # len(word) == 0
    if not word:
        return None
    if word[-1] == '"' and word[0] == '"':
        return word[1:-1]

    if(word == "None"):
        return None

    # parse uncertainty value
    uncertainty_rep = word.split("+-")
    if len(uncertainty_rep) == 2:
        try:
            ret = ufloat(float(uncertainty_rep[0]), float(uncertainty_rep[1]))
            return ret
        except ValueError:
            return word

    try:
        ret = float(word)
        return float(ret)
    except ValueError:
        return word


def mean_values(table):
    """Returns the mean values and errors of the given tables as ufloats."""

    _raw_dict = raw_data(table)
    mean_values = []
    for v in _raw_dict:
        mean_values.append(ufloat(np.mean(v), np.std(v)/np.sqrt(len(v))))

    return mean_values


def mean_values_dict(table):
    """Returns the mean values and errors of the given tables as ufloats."""

    _raw_dict = raw_dict(table)
    mean_values = {}
    for v in _raw_dict.items():
        mean_values[v[0]].append(np.mean(ufloat(np.mean(v[1]), np.std(v[1]))))

    return mean_values


def dict_from_table(table):
    """Retuns a dictionary mapping the first line to its columns"""
    dict_ret = {}
    for _c in table:
        dict_ret[_c[0]] = _c[1:]
    return dict_ret


def raw_dict(table):
    """Returns a dictionary mapping the first line to its non-string collumn values"""
    dictRet = {}
    for c in table:
        dictRet[c[0]] = []
        for j in c[1:]:
            # Use isinstance to account for numpy floats.
            if isinstance(j, numbers.Number) or isinstance(j, UFloat):
                dictRet[c[0]].append(j)

    return dictRet


def transposed(table):
    """Returns the transposition of the table."""
    if table == None:
        return None

    t_table = []
    for i in table:
        while len(i) > len(t_table):
            t_table.append([])

    for column in table:
        for i in range(len(column)):
            t_table[i].append(column[i])

    return t_table


def raw_data(table):
    """Returns numbers only"""

    data = []
    for i in table:
        data.append([])

    for i in range(len(table)):
        for w in table[i]:
            if isinstance(w, numbers.Number):
                data[i].append(w)
    return data


def gen_tex_table(
        table, filename,
        tex_caption="", tex_label="",
        subtables=None, precision="",
        midrule=2):
    """Generates a .tex file containing only a table.
    The whole file does not have to be modified anymore,
    it can be directly included with \\input{}.
    Uses LaTeX packages (include them!):
    * \\usepackage[
            locale=DE,
            separate-uncertainty=true,  % use \pm
            per-mode=symbol-or-fraction,
            %per-mode=reciprocal,
            %output-decimal-marker=.,
        ]{siunitx}
    * \\susepackage{subcaption}
    Uses S
    Parameters
    ----------
    table : [][] list of lists
        The table which should be parsed.
    ===
    filename : str
        The (relative) filepath where the table is saved.
    ===
    tex_caption : str
        The caption of the table in LaTeX.
    ===
    tex_label : str
        The label of the table in LaTeX.
    ===
    subtables : int or None
    The number of subtables to split into. (0 or None, for no subtables)
    ===
    precision : int[] or int or str[] or str
        The precision per column (as a list) or the precision for every number.
        The format string per column (see https://docs.python.org/3/library/string.html#format-string-syntax)
        or the format string for every number.
    ===
    midrule : int
        After which row to put the midrule. Every row after that will be treated as the header.
    """
    # Track row size to make it rectangular
    rows = 0
    # get the column meta data an put that in the list as __ColumnMeta
    column_meta = []
    # Gather Meta information and parse strings which can be parsed
    for col_index, column  in enumerate(table):
        meta_data = __ColumnMeta()
        if isinstance(precision, list):
            meta_data.precision = precision[col_index]
        else:
            meta_data.precision = precision
        if len(column) > rows:
            rows = len(column)
        for row_index, cell in enumerate(column):
            if isinstance(cell, UFloat):
                meta_data.HAS_UFloats = True
                if cell.std_dev >= 1 and meta_data.MAX_MAGNITUDE < int(log10(cell.std_dev)) + 1:
                    meta_data.MAX_MAGNITUDE = int(log10(cell.std_dev)) + 1
                if cell.nominal_value >= 1 and meta_data.MAX_MAGNITUDE < int(log10(cell.nominal_value)) + 1:
                    meta_data.MAX_MAGNITUDE = int(log10(cell.nominal_value)) + 1
            elif isinstance(cell, numbers.Number):
                if cell >= 1 and meta_data.MAX_MAGNITUDE < int(log10(cell)) + 1:
                    meta_data.MAX_MAGNITUDE = int(log10(cell)) + 1
        column_meta.append(meta_data)

    # Generate empty string table, to maintain rectangular size
    str_table = [["" for j in table] for i in range(rows)]

    # Generate all the strings in the table
    for col_index, column  in enumerate(table):
        for row_index, cell in enumerate(column):
            new_str = __tex_format__(cell, column_meta, col_index)
            if column_meta[col_index].MAX_LEN < len(new_str):
                column_meta[col_index].MAX_LEN = len(new_str)
            str_table[row_index][col_index] = new_str

    # Make the file
    with open(filename, "w") as file:
        file.write(r"\begin{table}"+"\n")
        file.write("\t"+r"\caption{"+tex_caption+"}\n")
        file.write("\t"+r"\label{"  +tex_label+"}\n")
        file.write("\t"+r"\centering"+"\n")
        # Make subtables
        if subtables and subtables != 0:
            if not isinstance(subtables, int):
                raise ValueError("Subtables must be None or a not negative integer.")
            header = str_table[:midrule]
            table = str_table[midrule:]
            rows_st = rows - midrule
            rows_per_subtable = math.ceil(rows_st/subtables)
            for i in range(subtables-1):
                file.write("\t"+r"\begin{subtable}{"+str(round(1.0/subtables - 0.01, 2))+r"\textwidth"+"}\n")
                file.write("\t\t"+r"\centering"+"\n")
                __write_tabular__(file, header+table[i*rows_per_subtable:(i+1)*rows_per_subtable]
                                  , column_meta, midrule, level=2)
                file.write("\t"+r"\end{subtable}"+"\n")
            # Write the last subtable
            file.write("\t"+r"\begin{subtable}{"+str(round(1.0/subtables - 0.01, 2))+r"\textwidth"+"}\n")
            file.write("\t\t"+r"\centering"+"\n")
            __write_tabular__(file, header+table[rows_per_subtable*(subtables-1):]
                                , column_meta, midrule, level=2, fillTo=midrule+rows_per_subtable)
            file.write("\t"+r"\end{subtable}"+"\n")

        else:
            __write_tabular__(file, str_table, column_meta, midrule)

        file.write(r"\end{table}")
        file.close()


def __write_tabular__(file, str_table, column_meta, midrule, level=1, fillTo=None):
    for _ in range(level):
        file.write("\t")
    file.write(r"\begin{tabular}{"+__si_table_header__(column_meta)+"}\n")
    for i in range(level):
        file.write("\t")
    file.write("\t"+r"\toprule"+"\n")
    for row_index, row in enumerate(str_table):
        for i in range(level):
            file.write("\t")
        file.write("\t")
        for col_index, cell in enumerate(row[:-1]):
            file.write(__tex_cell__(cell, column_meta[col_index].MAX_LEN))
            file.write(" & ")
        file.write(__tex_cell__(row[-1], column_meta[-1].MAX_LEN))
        file.write(r"\\"+"\n")
        if row_index + 1 == midrule:
            for i in range(level):
                file.write("\t")
            file.write("\t")
            file.write(r"\midrule"+"\n")
    if fillTo is not None and fillTo > len(str_table):
        for i in range(fillTo - len(str_table)):
            for i in range(level):
                file.write("\t")
            file.write("\t")
            for col_index, cell in enumerate(str_table[-1][:-1]):
                file.write(
                    r"\phantom{"+__tex_cell__(cell, column_meta[col_index].MAX_LEN).strip()+"}")
                file.write(" & ")
            file.write(
                r"\phantom{"+__tex_cell__(row[-1], column_meta[-1].MAX_LEN).strip()+"}")
            file.write(r"\\"+"\n")

    for i in range(level):
        file.write("\t")
    file.write("\t"+r"\bottomrule"+"\n")
    for i in range(level):
        file.write("\t")
    file.write(r"\end{tabular}"+"\n")


class __ColumnMeta:
    MAX_LEN = 0
    HAS_UFloats = False
    MAX_MAGNITUDE = 1
    precision = None


def __si_table_header__(column_meta):
    ret = ""
    for col in column_meta:
        if isinstance(col.precision, str):
            ret += r"S[table-format=" + col.precision + r"] "
        else:
            ret += r"S[table-format=" + \
                str(col.MAX_MAGNITUDE)+"."+str(col.precision) + r"] "
        if col.HAS_UFloats:
            ret += r"@{${}\pm{}$} "
            if isinstance(col.precision, str):
                ret += r"S[table-format=" + col.precision + r"] "
            else:
                ret += r"S[table-format=" + \
                    str(col.MAX_MAGNITUDE)+"."+str(col.precision) + r"] "
    return ret


def __tex_cell__(cell, max_len_column):
    remaining_spaces = max_len_column - len(cell)
    while remaining_spaces > 0:
        cell += " "
        remaining_spaces -= 1
    return cell


def __tex_format__(cell, column_meta, column_index):

    formatter = column_meta[column_index].precision

    # Numbers
    if isinstance(cell, numbers.Number):
        if isinstance(formatter, int):
            if formatter < 0:
                raise ValueError(
                    "Negative precision ("+formatter+") makes no sense.")
            formatter = "{:."+str(formatter)+"f}"

        elif isinstance(formatter, float):
            if formatter < 0:
                raise ValueError(
                    "Negative precision ("+formatter+") makes no sense.")
            formatter = ("{:" + str(int(formatter)) + "."
                         + str(int(round((formatter - int(formatter))*10))) + "f}"
                         )

        elif isinstance(formatter, str):
            formatter = "{:"+formatter+"f}"

        if column_meta[column_index].HAS_UFloats:
            warnings.warn(
                "You have normal numbers in a column where one or more UFloats exist. That can lead to ugly tables.")
            formatter += "&0"

        assert isinstance(formatter, str)

        return formatter.format(cell)

    if isinstance(cell, UFloat):
        if isinstance(formatter, str):
            cell_ufloat = ("{:"+formatter+"u}").format(cell).split("+/-")
        else:
            # Let uncertainty handle the precision
            cell_ufloat = "{:u}".format(cell).split("+/-")

        return cell_ufloat[0]+" & "+cell_ufloat[1]
    # cell is Null
    if not cell:
        return "{}"

    # No known rules (Maybe it is a string)
    # Add multicol stuff
    if column_meta[column_index].HAS_UFloats:
        return r"\multicolumn{2}{c}{"+str(cell)+r"}"
    return r"{"+str(cell)+r"}"



# standard imports
import sys
import numpy as np

from .tablehandler import TableHandler, genfromtxt

__all__ = []

if "matplotlib" in sys.modules or "matplotlib.pyplot" in sys.modules:
    print("Praktipy will not set up the matplotlib backend, because it is already set up.")
else:
    from .praktiplot import plt, matplotlib, cla
    __all__ += ["plt", "matplotlib", "cla"]

from scipy.optimize import curve_fit
from scipy import constants as const
from pint import UnitRegistry
from uncertainties import ufloat
import uncertainties.unumpy as unp


unit = UnitRegistry()
noms = unp.nominal_values
stds = unp.std_devs

# standart import useful (SI) units
meter = unit.meter
seconds = unit.seconds
kilogram = unit.kilogram
kelvin = unit.kelvin
celsius = unit.celsius


# convenience functions
def polyplotfit(x, params, N=1000, border=0.05):
    """Plots a polynome, which was fitted.
    x: the original x value which was fitted
    params: the parameters in the polynome
    N: Number of x_values to calculate
    border: percentage of x_range to make a border"""

    dx = x[-1] - x[0]
    x_fit = np.linspace(x[0] - dx*border, x[-1] + dx*border, N)
    y_fit = np.zeros(len(x_fit))
    deg = len(params)
    for i in range(deg):
        y_fit += params[deg-1-i] * x_fit**i

    return (x_fit, y_fit)


def curveplotfit(f, x, params, N=1000, border=0.05, logscale=False):
    """Plots a general function, which was fitted.
    x: the original x value which was fitted
    params: the parameters in the polynome
    N: Number of x_values to calculate
    border: percentage of x_range to make a border"""
    dx = x[-1] - x[0]
    if logscale:
        x_fit = np.logspace(np.log10(
            x[0]) - np.log10(x[0])*border*0.6, np.log10(x[-1])+np.log10(x[-1])*border*0.2, N)
    else:
        x_fit = np.linspace(x[0] - dx*border, x[-1] + dx*border, N)
    y_fit = f(x_fit, *params)

    return (x_fit, y_fit)


units = ["meter", "seconds", "kilogram", "kelvin", "celsius"]

__all__ += ["TableHandler", "np", "ufloat", "unp", "unit", "noms",
            "stds", "const", "polyplotfit", "genfromtxt", "curveplotfit", "curve_fit"]
__all__ += units



# standard imports
import sys
import numpy as np

from . import tablehandler as th
__all__ = []

if "matplotlib" in sys.modules or "matplotlib.pyplot" in sys.modules:
    print("Praktipy will not set up the matplotlib backend, because it is already set up.")
else:
    from .praktiplot import plt, matplotlib
    from . import praktiplot as pplt
    __all__ += ["plt", "matplotlib", "pplt"]

from scipy.optimize import curve_fit
from scipy import constants as const
from pint import UnitRegistry
from uncertainties import ufloat
import uncertainties.unumpy as unp

unit = UnitRegistry()
noms = unp.nominal_values
stds = unp.std_devs

# standard import useful (SI) units
meter = unit.meter
seconds = unit.seconds
kilogram = unit.kilogram
kelvin = unit.kelvin
celsius = unit.celsius

units = ["meter", "seconds", "kilogram", "kelvin", "celsius"]

__all__ += ["th", "np", "ufloat", "unp", "unit", "noms",
            "stds", "const", "curve_fit"]
__all__ += units



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'praktipy'
DESCRIPTION = 'A small toolbox for physics laboratory courses at the TU Dortmund.'
URL = 'https://github.com/The-Ludwig/praktipy'
AUTHOR = 'Ludwig Neste <The-Ludwig>, Max Uetrecht <phenomax>'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

REQUIRED = [
    'matplotlib', 'uncertainties', 'numpy', 'pint', 'scipy'
]

DEV_REQUIRED = [
    'twine'
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    extras_require={
        'dev': DEV_REQUIRED,
    },
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Natural Language :: German',
        'Operating System :: OS Independent',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Education'
    ],
)
